/////////////////////////////////////////////////
//
// DrBinCas | A small utility for converting
//            binary files to Dragon .CAS files.
//
// Version 0.1
//
// By Robcfg
//
// The program returns one of the following:
//   -1, if the file could not be loaded or read,
//       or if program arguments weren't right.
//    0, if conversion went Ok.
//
/////////////////////////////////////////////////

#define _CRT_SECURE_NO_WARNINGS // For Microsoft Visual Studio

#include <iostream>
#include <string>
#include <stdio.h>

using namespace std;

#define CAS_NAME_LEN        8
#define CAS_PILOT_VAL       0x55
#define CAS_BLOCK_END       0x55
#define CAS_BLOCK_START     0x3C
#define CAS_BLOCK_NAME      0x00
#define CAS_BLOCK_NAME_LEN  0x0F
#define CAS_BLOCK_DATA      0x01
#define CAS_BLOCK_EOF       0xFF

// Computes a block CRC.
// The CRC is the sum of all the data values in the block
// plus the block type and block size.
unsigned char GetBlockCRC( const unsigned char* pData, unsigned char len, unsigned char blockType )
{
  unsigned char retVal = 0;

  for( unsigned int i = 0; i < len; ++i )
  {
    retVal += pData[i];
  }
  retVal += blockType;
  retVal += len;

  return retVal;
}

// Processes the file data and writes it in Dragon .CAS format
int SaveCasFile( const char* name, FILE* inFile, FILE* outFile, unsigned short int loadAdd, unsigned short int execAdd )
{
  unsigned char pilot = CAS_PILOT_VAL;

  // Process name
  char finalName[CAS_NAME_LEN];
  memset( finalName, 0x20, CAS_NAME_LEN );
  unsigned int nameLen = (unsigned int)strlen(name);

  if( nameLen >= CAS_NAME_LEN )
  {
    nameLen = CAS_NAME_LEN;
  }

  memcpy( finalName, name, nameLen );

  // Write Leader block
  for( int i = 0; i < 16; ++i )
  {
    fwrite( &pilot, 1, 1, outFile );
  }

  // Write Name block
  unsigned char* nameBlock = new unsigned char[20];
  unsigned char* pLoadAdd = (unsigned char*)&loadAdd;
  unsigned char* pExecAdd = (unsigned char*)&execAdd;

  nameBlock[0] = CAS_BLOCK_START;
  nameBlock[1] = CAS_BLOCK_NAME;
  nameBlock[2] = CAS_BLOCK_NAME_LEN;
  memcpy( &nameBlock[3], finalName, CAS_NAME_LEN );
  nameBlock[11] = 0x02; // Binary file
  nameBlock[12] = 0x00; // Binary file
  nameBlock[13] = 0x00; // Continuous data stream for binary files
  nameBlock[14] = pLoadAdd[1];
  nameBlock[15] = pLoadAdd[0];
  nameBlock[16] = pExecAdd[1];
  nameBlock[17] = pExecAdd[0];
  nameBlock[18] = GetBlockCRC( &nameBlock[3], nameBlock[2], nameBlock[1] );
  nameBlock[19] = CAS_BLOCK_END;

  fwrite( nameBlock, 1, 20, outFile );

  delete nameBlock;

  // Write binary data blocks
  unsigned char* dataBlock = new unsigned char[260];
  unsigned int uRead = (unsigned int)fread( &dataBlock[3], 1, 0xFF, inFile);

  while( uRead != 0x0 )
  {
    dataBlock[0] = CAS_BLOCK_START;
    dataBlock[1] = CAS_BLOCK_DATA;
    dataBlock[2] = (unsigned char)uRead;
    dataBlock[3 + uRead] = GetBlockCRC( &dataBlock[3], dataBlock[2], dataBlock[1] );
    dataBlock[4 + uRead] = CAS_BLOCK_END;

    fwrite( nameBlock, 1, uRead+5, outFile );

    uRead = (unsigned int)fread( &dataBlock[3], 1, 0xFF, inFile);
  }

  delete dataBlock;

  // Write End-Of-File Block
  unsigned char* endBlock = new unsigned char[5];

  endBlock[0] = CAS_BLOCK_START;
  endBlock[1] = CAS_BLOCK_EOF;
  endBlock[2] = 0;
  endBlock[3] = CAS_BLOCK_EOF; // In this case, as there is no data, the CRC equals the block type.
  endBlock[4] = CAS_BLOCK_END;

  fwrite( endBlock, 1, 5, outFile );

  delete endBlock;

  return 0;
}

int main(int argc, char* argv[])
{
  // Variables
  FILE* inFile = 0;
  FILE* outFile = 0;
  unsigned char* pData = 0;

  // Print banner
  cout << endl;
  cout << "DrBinCas v0.1 | Program for converting binary files to .CAS files by Robcfg" << endl;
  cout << "----------------------------------------------------------------------------" << endl << endl;

  // Print arguments (for debug purposes only)
#ifdef _DEBUG
  for( int iArg = 0; iArg < argc; ++iArg )
  {
    cout << argv[iArg] << endl;
  }
  cout << endl;
#endif

  // Check arguments
  if( argc < 5 )
  {
    cout << "You didn't specify all required arguments." << endl << endl;
    cout << "Usage:" << endl << endl;
    cout << "DrBinCas filename program_name load_address exec_address" << endl << endl;
    cout << "For example: DrBinCas \"d:\\my files\\example.bin\" EXAMPLE 16384 16384" << endl << endl;
    return -1;
  }

  if( argc > 5 )
  {
    cout << "Too many arguments. Please check the file name for spaces." << "In case the file name contains spaces, wrap it with double quotes." << endl << endl;
    cout << "For example: DrBinCas \"d:\\my files\\example.bin\" EXAMPLE 16384 16384" << endl << endl;
    return -1;
  }

  // Open input file
  char* fileName = argv[1];

  inFile = fopen( fileName, "rb" );
  if( NULL == inFile )
  {
    cout << "Could not open " << fileName << " for reading." << endl;
    cout << "Please check the file name and make sure that the file is not in use." << endl << endl;
    return -1;
  }

  // Open output file
  string outFileName = fileName;
  outFileName += ".cas";

  outFile = fopen( outFileName.c_str(), "wb" );
  if( NULL == outFile )
  {
    cout << "Could not open " << outFileName.c_str() << " for writing." << endl;
    cout << "Please check the file name and make sure that the file is not in use." << endl << endl;
    return -1;
  }

  // Save data in .cas format
  unsigned short int loadAdd = (unsigned short int)atoi( argv[3] );
  unsigned short int execAdd = (unsigned short int)atoi( argv[4] );

  SaveCasFile( argv[2], inFile, outFile, loadAdd, execAdd );

  // Close files
  fclose( outFile );
  fclose( inFile  );
  
  return 0;
}